/*----------------------------------------------------------------------------
/ M-800PIO												2006 (C)LogicPack
/	06.02.25									000		M.Yamada
/		first release
/	06.05.01	V1.00									M.Yamada
/---------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "iodefine.h"
#include "sci.h"
#include "m802.h"
#include "command.h"
#include "debug.h"
#include "mstring.h"

#ifdef __cplusplus
extern "C" {
#endif
void abort(void);
#ifdef __cplusplus
}
#endif

/*-----------------------------------------------
/ *GLOBAL
/----------------------------------------------*/
#pragma section EXT
volatile long	tim1ms;		/* ėp^C}JE^(msec) */
#pragma section

/*-----------------------------------------------
/ *DEBUG
/----------------------------------------------*/
//#define DEBUG_DUMMY_PORT

/*-----------------------------------------------
/ *DEFINE
/----------------------------------------------*/
#define CS3_ADDR		0x600000
#define CS4_ADDR		0x800000
#define CS5_ADDR		0xA00000
#define CS6_ADDR		0xC00000
#define CS7_ADDR		0xE00000

#define M800IO_AOFS		0x2000

#define PA_OFS			0x00
#define PB_OFS			0x02
#define PC_OFS			0x04
#define PD_OFS			0x06
#define PE_OFS			0x08
#define PF_OFS			0x0A
//#define DIR_OFS			0x10						//DEL
#define DIRE_OFS		0x18
#define DIRF_OFS		0x1A

#if !defined(DEBUG_DUMMY_PORT)
  #define M800IO_PA		(*(volatile byte *)(m800io_config.addr + PA_OFS))
  #define M800IO_PB		(*(volatile byte *)(m800io_config.addr + PB_OFS))
  #define M800IO_PC		(*(volatile byte *)(m800io_config.addr + PC_OFS))
  #define M800IO_PD		(*(volatile byte *)(m800io_config.addr + PD_OFS))
  #define M800IO_PE		(*(volatile byte *)(m800io_config.addr + PE_OFS))
  #define M800IO_PF		(*(volatile byte *)(m800io_config.addr + PF_OFS))
//  #define M800IO_DIR	(*(volatile byte *)(m800io_config.addr + DIR_OFS))	//DEL
  #define M800IO_DIRE	(*(volatile byte *)(m800io_config.addr + DIRE_OFS))
  #define M800IO_DIRF	(*(volatile byte *)(m800io_config.addr + DIRF_OFS))
#else
  static byte dummyp[4];
  #define M800IO_PA		dummyp[0]
  #define M800IO_PB		dummyp[1]
  #define M800IO_PC		dummyp[2]
  #define M800IO_PD		dummyp[0]
  #define M800IO_PE		dummyp[1]
  #define M800IO_PF		dummyp[2]
  #define M800IO_DIR	dummyp[3]
#endif

typedef struct{
	int cs;
	int bid;
	dword addr;
}M800IO_CONFIG;

/*-----------------------------------------------
/ *STATIC
/----------------------------------------------*/
/* message string */
const char prg[] = "M-800PIO test V1.00";

static M800IO_CONFIG m800io_config = {
	3, 0, CS3_ADDR + M800IO_AOFS + 0
};

/* prototype */
/*-----*/
static void disp_title(void);			/* ^Cg\ */
static void disp_boardinfo(void);
/*-----*/

void debug_out_int(const char *top, int data, const char *bottom);
void con_puts(const char *str);
void waitms(int ms);

bool cmd_help(const PARAM_INFO *param);
bool cmd_config(const PARAM_INFO *param);
bool cmd_direction(const PARAM_INFO *param);
bool cmd_input(const PARAM_INFO *param);
bool cmd_output(const PARAM_INFO *param);
bool cmd_inputlog(const PARAM_INFO *param);
bool cmd_test(const PARAM_INFO *param);

static void outp(int c, byte data);
static byte inp(int c);

/*-------------------------------------------------------------------
/ MAIN
/------------------------------------------------------------------*/
void main(void)
{
	long time;

	sci1_init();
	disp_title();
	disp_boardinfo();

	time = tim1ms;

	cmd_prompt("m800pio>");
	/* R}h̓o^ */
	cmd_entry("Set",	1, cmd_config);
	cmd_entry("Dir",	1, cmd_direction);
	cmd_entry("In",		1, cmd_input);
	cmd_entry("Out",	1, cmd_output);
	cmd_entry("Log",	1, cmd_inputlog);
	cmd_entry("Test",	1, cmd_test);
	cmd_entry("?",		1, cmd_help);
	cmd_entry("Help",	1, cmd_help);

	while(1){
		cmd_job();
	}
}

/*-------------------------------------------------------------------
/ abort
/------------------------------------------------------------------*/
void abort(void)
{
}

/*---------------------------------------------------------
/ ^Cg\
/--------------------------------------------------------*/
static void disp_title(void)
{
	const char waku[] = 
		"-----------------------------------------------------------\r\n";
#if defined(RAM_EMULATE)
	waitms(2000);
#endif

	debug_out(waku);
	debug_out("--- ");	debug_out(prg);
#if defined(RAM_EMULATE)
	debug_out(" (ram emulation)\r\n");
#else
	debug_out("\r\n");
#endif
	debug_out(waku);
}

/*---------------------------------------------------------
/ ^Cg\
/--------------------------------------------------------*/
static void disp_boardinfo(void)
{
	const char str[] = 
		" Port A,B    Isorate Input\r\n"
		" Port C,D    Isorate Output\r\n"
		" Port E,F    I/O(3.3V, bit direction)\r\n"
		"-----------------------------------------------------------\r\n";
	debug_out(str);
}

/* R}h^Cg\ */
static void title_cmd(const char *str)
{
	debug_out("\r\n*");
	debug_out(str);
	debug_out("\r\n");
}

/*---------------------------------------------------------
/ R}hʃp[^́Zbg
/--------------------------------------------------------*/
/* Help */
/* ?    */
bool cmd_help(const PARAM_INFO *param)
{
	const char help[] = 
		" Set <cs> <bid>       Board Configration\r\n"
		"                      cs  : chip select number(default:3)\r\n"
		"                      bid : BOARD ID(default:0)\r\n"
		" Dir <port> <data>     Port Input/Output Select\r\n"
		"                      port : E,F\r\n"
		"                      data : Direction\r\n"
		"                      ex) Dir E f0\r\n"
		" Input <port>         Display data from Input <port>\r\n"
		" Output <port> <data> Output <data> to <port>\r\n"
		" Log [msec] [S]       Log Input data\r\n"
		"                      msec : sampling time(default:200msec, minimax:100msec)\r\n"
		"                      S    : Scroll mode\r\n"
		" Test                 Loopback test\r\n"
		" Help or ?            disp help\r\n"
		"\r\n";
	con_out(help);
	return TRUE;
}

/* Board Configration  */
/* Set <cs> <bid> */
bool cmd_config(const PARAM_INFO *param)
{
	dword addr;
	int bid;
	int cs;

	title_cmd("Board Configration");

	if( param->count != 2 ){ goto PDISP; }

	cs = str2dec(param->p[0]);
	switch( cs ){
		case 3: addr = CS3_ADDR + M800IO_AOFS; break;
		case 4: addr = CS4_ADDR + M800IO_AOFS; break;
		case 5: addr = CS5_ADDR + M800IO_AOFS; break;
		case 6: addr = CS6_ADDR + M800IO_AOFS; break;
		case 7: addr = CS7_ADDR + M800IO_AOFS; break;
		default: goto PERROR;
	}
	bid = str2dec(param->p[1]);
	if( bid > 15 ){
PERROR:
		con_out("parameter error\r\n");
		return TRUE;
	}

	addr += (bid << 8);
	m800io_config.addr = addr;
	m800io_config.cs = cs;
	m800io_config.bid = bid;
PDISP:
	debug_out_int("M800IO access to /CS", m800io_config.cs, "\r\n");
	debug_out_int("M800IO BOARD ID=", m800io_config.bid, "\r\n\r\n");
	return TRUE;
}

#define PA_DIR			0x01
#define PB_DIR			0x02
#define PC_DIR			0x04
#define PD_DIR			0x08
#define PE_DIR			0x10
#define PF_DIR			0x20
const byte sel_tbl[] = { PA_DIR, PB_DIR, PC_DIR, PD_DIR, PE_DIR, PF_DIR };

/* Port Direction */
/* Dir <port> <data> */
bool cmd_direction(const PARAM_INFO *param)
{
	#define PE_DIR_BIT	0x01
	#define PF_DIR_BIT	0x02
	const byte dirtbl[] = { PE_DIR_BIT, PF_DIR_BIT };
	const char *p;
	int pc;
	byte bit = 0;
	byte dir;
	char *pp;
	int i;

	title_cmd("Port Direction");

	if( param->count != 2 ){
		debug_out("bit  7   6   5   4   3   2   1   0\r\n");
		dir = M800IO_DIRE;
		debug_out("PE: ");
		for( i = 0; i < 8; i++, dir <<= 1 ){
			if( dir & 0x80 ){
				debug_out("Out ");
			}else{
				debug_out("In  ");
			}
		}
		debug_out("\r\n");
		dir = M800IO_DIRF;
		debug_out("PF: ");
		for( i = 0; i < 8; i++, dir <<= 1 ){
			if( dir & 0x80 ){
				debug_out("Out ");
			}else{
				debug_out("In  ");
			}
		}
		debug_out("\r\n\r\n");
		return TRUE;
	}
	
	for( pp = param->p[0]; *pp != '\0'; pp++ ){
		/* Port */
		pc = toupper(*pp);
		if( pc != 'E' && pc != 'F' ){
PERROR:
			con_out("parameter error\r\n");
			return TRUE;
		}
		pc -= 'E';
		bit |= dirtbl[pc];
	}
	/* I or O */
	dir = (byte)str2hex(param->p[1]);
	if( bit & PE_DIR_BIT ){
		M800IO_DIRE = dir;
		debug_outb3("Set PE direction:", dir, "\r\n");
	}
	if( bit & PF_DIR_BIT ){
		M800IO_DIRF = dir;
		debug_outb3("Set PF direction:", dir, "\r\n");
	}
	debug_out("\r\n");
	return TRUE;
}

/* Port Input */
/* In <port> */
bool cmd_input(const PARAM_INFO *param)
{
	const char *port_str[] = { "PA", "PB", "PC", "PD", "PE", "PF" };
	const char *p;
	int pc;
	byte data[6];
	char *pp;
	byte bit = 0;
	int i;
	bool cntf = FALSE;

	title_cmd("Port Input");

	if( param->count != 1 ){ goto PERROR; }

	for( pp = param->p[0]; *pp != '\0'; pp++ ){
		/* Port */
		pc = toupper(*pp);
		if( !(pc >= 'A' && pc <= 'F') ){
PERROR:
			con_out("parameter error\r\n");
			return TRUE;
		}
		pc -= 'A';
		bit |= sel_tbl[pc];
	}
	/* input */
	for( i = 0; i < 6; i++ ){
		if( bit & sel_tbl[i] ){
			data[i] = inp(i + 'A');
		}
	}
	/* display */
	debug_out("Input ");
	for( i = 0; i < 6; i++ ){
		if( bit & sel_tbl[i] ){
			if( cntf ){ debug_out(", "); }
			debug_out(port_str[i]);
			debug_outb3(":", data[i], "");
			cntf = TRUE;
		}
	}
	debug_out("\r\n\r\n");

	return TRUE;
}

/* Port Output */
/* Out <port> <data> */
bool cmd_output(const PARAM_INFO *param)
{
	const char *port_str[] = { "PA", "PB", "PC", "PD", "PE", "PF" };
	const char *p;
	int pc;
	byte data;
	char *pp;
	byte bit = 0;
	int i;
	bool cntf = FALSE;

	title_cmd("Port Output");

	if( param->count != 2 ){ goto PERROR; }

	for( pp = param->p[0]; *pp != '\0'; pp++ ){
		/* Port */
		pc = toupper(*pp);
		if( !(pc >= 'A' && pc <= 'F') ){
PERROR:
			con_out("parameter error\r\n");
			return TRUE;
		}
		pc -= 'A';
		bit |= sel_tbl[pc];
	}
	/* data */
	data = (byte)str2hex(param->p[1]);
	/* output */
	for( i = 0; i < 6; i++ ){
		if( bit & sel_tbl[i] ){
			outp(i + 'A', data);
		}
	}
	/* display */
	debug_out("Output ");
	for( i = 0; i < 6; i++ ){
		if( bit & sel_tbl[i] ){
			if( cntf ){ debug_out(", "); }
			debug_out(port_str[i]);
			debug_outb3(":", data, "");
			cntf = TRUE;
		}
	}
	debug_out("\r\n\r\n");

	return TRUE;
}

/* ͊Ď */
/* Log [msec] [S] */
bool cmd_inputlog(const PARAM_INFO *param)
{
	static byte pa, pb, pc, pd, pe, pf;
	long stime = 200;
	long time;
	int i;
	bool scroll = FALSE;

	title_cmd("Input Port Log");

	if( param->count > 2 ){
PERROR:
		con_out("parameter error\r\n");
		return TRUE;
	}

	for( i = 0; i < param->count; i++ ){
		if( toupper(*(param->p[i])) == 'S' ){
			scroll = TRUE;				/* XN[[h */
		}else{
			long msec;
			msec = str2dec(param->p[i]);
			if( msec < 100 ){ goto PERROR; }
			stime = msec;
		}
	}

	time = tim1ms;

	while( !(sci1_get() & 0x100) ){
		if( tim1ms - time >= stime ){
			time = tim1ms;
			pa = inp('A');	pb = inp('B');	pc = inp('C');
			pd = inp('D');	pe = inp('E');	pf = inp('F');
			debug_outb3("PA:", pa, " ");	debug_outb3("PB:", pb, " ");
			debug_outb3("PC:", pc, " ");	debug_outb3("PD:", pd, " ");
			debug_outb3("PE:", pe, " ");	debug_outb3("PF:", pf, " ");
			if( scroll ){
				debug_out("\r\n");
			}else{
				debug_out("\r");
			}
		}
	}
	debug_out("\r\n\r\n");

	return TRUE;
}

/* Loopback Test */
/* Test */

#define ABC_DIR		(PA_DIR | PB_DIR | PC_DIR)
#define DEF_DIR		(PD_DIR | PE_DIR | PF_DIR)


#define TEST1_MAX	18
#define TEST1_PMAX	2

typedef struct{
	char		*omsg;
	byte		data[TEST1_PMAX];
}TEST_TABLE1;

const TEST_TABLE1 test_tbl1[TEST1_MAX] = {
	{ "CD ALL OFF", {0x00, 0x00} },
	{ "PC0 ON", {0x01, 0x00} },
	{ "PC1 ON", {0x02, 0x00} },
	{ "PC2 ON", {0x04, 0x00} },
	{ "PC3 ON", {0x08, 0x00} },
	{ "PC4 ON", {0x10, 0x00} },
	{ "PC5 ON", {0x20, 0x00} },
	{ "PC6 ON", {0x40, 0x00} },
	{ "PC7 ON", {0x80, 0x00} },
	{ "PD0 ON", {0x00, 0x01} },
	{ "PD1 ON", {0x00, 0x02} },
	{ "PD2 ON", {0x00, 0x04} },
	{ "PD3 ON", {0x00, 0x08} },
	{ "PD4 ON", {0x00, 0x10} },
	{ "PD5 ON", {0x00, 0x20} },
	{ "PD6 ON", {0x00, 0x40} },
	{ "PD7 ON", {0x00, 0x80} },
	{ "CD ALL ON", {0xff, 0xff} }
};

bool cmd_test(const PARAM_INFO *param)
{
	const char *pab[2] = { "PA", "PB" };
	const char _cd[] = "CD";
	const char _ab[] = "AB";
	const TEST_TABLE1 *tp;
	int i, j;
	static byte rdt[TEST1_MAX];
	#define rdata rdt[0]
	#define wdata rdt[1]

	title_cmd("Loopback Test");

	for( i = 0; i <TEST1_MAX; i++ ){
		tp = &test_tbl1[i];
		/*--- o ---*/
		debug_out(tp->omsg);	debug_out(" -> ");
		for( j = 0; j < TEST1_PMAX; j++ ){
			outp(_cd[j], tp->data[j]);
		}
		waitms(5);
		/*---  ---*/
		for( j = 0; j < TEST1_PMAX; j++ ){
			rdt[j] = inp(_ab[j]);
			debug_out(pab[j]);
			debug_outb3(":", rdt[j], " ");
		}
		/*---  ---*/
		for( j = 0; j < TEST1_PMAX; j++ ){
			if( rdt[j] != tp->data[j] ){ break; }	/* error */
		}
		if( j == TEST1_PMAX ){
			debug_out("OK\r\n");
		}else{
			debug_out("NG\r\n");
		}
	}
	/* Port E -> F */
	M800IO_DIRF = 0x00;
	M800IO_DIRE = 0xff;
	debug_out("PE ALL LOW -> ");
	wdata = 0x00;
	outp('E', wdata);
	rdata = inp('F');
	debug_outb3("PF:", rdata, " ");
	if( wdata == rdata ){
		debug_out("OK\r\n");
	}else{
		debug_out("NG\r\n");
	}
	for( i = 0; i < 8; i++ ){
		debug_out_int("PE", i, " HIGH -> ");
		wdata = 0x01 << i;
		outp('E', wdata);
		rdata = inp('F');
		debug_outb3("PF:", rdata, " ");
		if( wdata == rdata ){
			debug_out("OK\r\n");
		}else{
			debug_out("NG\r\n");
		}
	}
	debug_out("PE ALL HIGH -> ");
	wdata = 0xff;
	outp('E', wdata);
	rdata = inp('F');
	debug_outb3("PF:", rdata, " ");
	if( wdata == rdata ){
		debug_out("OK\r\n");
	}else{
		debug_out("NG\r\n");
	}
	/* Port F -> E */
	M800IO_DIRE = 0x00;
	M800IO_DIRF = 0xff;
	debug_out("PF ALL LOW -> ");
	wdata = 0x00;
	outp('F', wdata);
	rdata = inp('E');
	debug_outb3("PE:", rdata, " ");
	if( wdata == rdata ){
		debug_out("OK\r\n");
	}else{
		debug_out("NG\r\n");
	}
	for( i = 0; i < 8; i++ ){
		debug_out_int("PF", i, " HIGH -> ");
		wdata = 0x01 << i;
		outp('F', wdata);
		rdata = inp('E');
		debug_outb3("PE:", rdata, " ");
		if( wdata == rdata ){
			debug_out("OK\r\n");
		}else{
			debug_out("NG\r\n");
		}
	}
	debug_out("PF ALL HIGH -> ");
	wdata = 0xff;
	outp('F', wdata);
	rdata = inp('F');
	debug_outb3("PE:", rdata, " ");
	if( wdata == rdata ){
		debug_out("OK\r\n");
	}else{
		debug_out("NG\r\n");
	}
	debug_out("\r\n");
	return TRUE;
}

static void outp(int c, byte data)
{
	switch( c ){
		case 'A': M800IO_PA = data; break;
		case 'B': M800IO_PB = data; break;
		case 'C': M800IO_PC = data; break;
		case 'D': M800IO_PD = data; break;
		case 'E': M800IO_PE = data; break;
		case 'F': M800IO_PF = data; break;
	}
}
static byte inp(int c)
{
	byte data = 0x00;
	switch( c ){
		case 'A': data = M800IO_PA; break;
		case 'B': data = M800IO_PB; break;
		case 'C': data = M800IO_PC; break;
		case 'D': data = M800IO_PD; break;
		case 'E': data = M800IO_PE; break;
		case 'F': data = M800IO_PF; break;
	}
	return data;
}

/*-------------------------------------------------------------------
/ int 10i\
/------------------------------------------------------------------*/
void debug_out_int(const char *top, int data, const char *bottom)
{
	static char tmp[10];
	debug_out(top);
	debug_out(utoa(data, tmp));
	debug_out(bottom);
}

/*-------------------------------------------------------------------
/ R\[o
/------------------------------------------------------------------*/
void con_puts(const char *str)
{
	con_out(str);
	con_out("\r\n");
}
/*-------------------------------------------------------------------
/ EFCgimsecj
/------------------------------------------------------------------*/
void waitms(int ms)
{
	long time;
	time = tim1ms;
	while(tim1ms - time <= ms){;}
}
